/***************************************************************************
                          game.c  -  description
                             -------------------
    begin                : St z 17 2003
    copyright            : (C) 2003 by Petr Cermak
    email                : cermak01@seznam.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "game.h"
#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <fcntl.h>

/**
 * Initializes/allocates the Game structure
 */
Game* newGame(int rows, int cols, GameState whoStarts, GameState whoAmI, int pisLen)
{
	int i, j;
	Game* g = (Game*) malloc(sizeof(Game));
	if (!g || pisLen > rows || pisLen > cols) {
		return NULL;
	}


	g->fields = (FieldState**) malloc(sizeof(FieldState*) * rows);
	for (i = 0; i < rows; i++) {
		g->fields[i] = (FieldState*) malloc(sizeof(FieldState) * cols);
		if (!g->fields[i]) {
			return NULL;
		}
	}

	for (i = 0; i < cols; i++) {
		for (j = 0; j < rows; j++) {
			g->fields[j][i] = FS_BLANK;
		}
	}

	g->rows = rows;
	g->cols = cols;
	g->empty = rows * cols;
	g->gs = whoStarts;
	g->whoAmI = whoAmI;
	g->redScore = 0;
	g->blueScore = 0;
	g->pisLen = pisLen;

	return g;
}

void deleteGame(Game* g)
{
	int i;

	for (i = 0; i < g->rows; i++) {
		free(g->fields[i]);
	}
	free(g->fields);
	free(g);
}


void oppositeWhoAmI(Game* g)
{
	if (g->whoAmI == GS_BLUE) {
		g->whoAmI = GS_RED;
	} else {
		g->whoAmI = GS_BLUE;
	}
}

void switchTurn(Game* g)
{
	if (g->gs == GS_BLUE) {
		g->gs = GS_RED;
	} else {
		g->gs = GS_BLUE;
	}
}

static FieldState getCheckedState(FieldState fs)
{
	switch (fs) {
		case FS_BLUE:
			return FS_BLUEX;
			break;
		case FS_RED:
			return FS_REDX;
			break;
		default:
			return fs;
			break;
	}
}

/**
 * Marks fields starting at startx, starty in direction given by dx, dy with new state fs.
 */
static void markChecked(Game* g, FieldState fs, int startx, int starty, int dx, int dy)
{
	int i, j;
	FieldState newFs;

	newFs = getCheckedState(fs);

	i = startx;
	j = starty;

	while (i >= 0 && j >= 0 && i < g->cols && j < g->rows && g->fields[j][i] == fs) {
		g->fields[j][i] = newFs;
		i += dx;
		j += dy;
	}
}

/**
 * Handles the move of player fs on position x, y
 *
 * @return number of _piskvorku_ created
 */
static int handleXY(Game* g, FieldState fs, int x, int y)
{
	int i, j;
	int len;
	int lastx, lasty, startx, starty;
	int pisCount = 0;


	// Check row
	g->fields[y][x] = fs;
	
	lastx = -2;
	startx = -2;
	len = 0;
	for (i = 0; i < g->cols; i++) {
		if (g->fields[y][i] == fs) {
			if (lastx < (i - 1)) {
				startx = lastx = i;
				len = 1;
			} else {
				lastx++;
				len++;
			}
			if (len >= g->pisLen) {
				markChecked(g, fs, startx, y, 1, 0);
				pisCount++;
				break;
			}
		}
	}

	// Check col
	g->fields[y][x] = fs;
	
	lasty = -2;
	starty = -2;
	len = 0;
	for (i = 0; i < g->rows; i++) {
		if (g->fields[i][x] == fs) {
			if (lasty < (i - 1)) {
				starty = lasty = i;
				len = 1;
			} else {
				len++;
				lasty++;
			}
			if (len >= g->pisLen) {
				markChecked(g, fs, x, starty, 0, 1);
				pisCount++;
				break;
			}
		}
	}

	// Check diag 1 1
	g->fields[y][x] = fs;
	
	if (x <= y) {
		i = 0;
		j = y - x;
	} else {
		i = x - y;
		j = 0;
	}

	len = 0;
	lastx = lasty = -2;
	startx = starty = -2;
	for (; (i < g->cols) && j < (g->rows); i++, j++) {
		if (g->fields[j][i] == fs) {
			if (lastx < (i - 1)) {
				starty = lasty = j;
				startx = lastx = i;
				len = 1;
			} else {
				lasty++;
				lastx++;
				len++;
			}
			if (len >= g->pisLen) {
				markChecked(g, fs, startx, starty, 1, 1);
				pisCount++;
				break;
			}
		}
	}

	// Check diag -1 1
	g->fields[y][x] = fs;
	
	if ((g->cols - x - 1) <= y) {
		i = g->cols - 1;
		j = y - (g->cols - x - 1);
	} else {
		i = x + y;
		j = 0;
	}

	len = 0;
	lastx = lasty = -2;
	startx = starty = -2;
	for (; i >= 0 && j < g->rows; i--, j++) {
		if (g->fields[j][i] == fs) {
			if (lasty < (j - 1)) {
				starty = lasty = j;
				startx = lastx = i;
				len = 1;
			} else {
				lastx--;
				lasty++;
				len++;
			}
			if (len == g->pisLen) {
				markChecked(g, fs, startx, starty, -1, 1);
				pisCount++;
				break;
			}
		}
	}

	if (pisCount > 0) {
		g->fields[y][x] = getCheckedState(fs);
	}

	return pisCount;

}

/**
 * Plays fs at position x, y
 */
Boolean playAt(Game* g, FieldState fs, int x, int y)
{
	int pisCount;
	
	if (x >= g->cols || y >= g->rows) {
		return FALSE;
	}

	if ((fs != FS_BLUE && fs != FS_RED) || (g->fields[y][x] != FS_BLANK)) {
		return FALSE;
	}

	pisCount = handleXY(g, fs, x, y);
	g->empty--;
	if (fs == FS_BLUE) {
		g->blueScore += pisCount;
	} else {
		g->redScore += pisCount;
	}
	return TRUE;
}

void pokus()
{
	void* a = malloc(1000);
	int fd;

	fd = open("/home/petr/prg/netpis/netpis/a.txt", O_RDONLY);
	read(fd, a, 1000);
	close(fd);
	fd = creat("/home/petr/prg/netpis/netpis/b.txt", 0666);
	write(fd, a, 1000);
	close(fd);
}
