/***************************************************************************
                          game.c  -  description
                             -------------------
    begin                : St z 17 2003
    copyright            : (C) 2003 by Petr Cermak
    email                : cermak01@seznam.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#define NOMACROS
#include <stdio.h>
#include <unistd.h>
#include <sys/select.h>
#include <string.h>
//#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <netdb.h>
#include <sys/poll.h>
#include <curses.h>
#include <pthread.h>
#include <signal.h>

#include "game.h"
#include "shared.h"
#include "net.h"

typedef enum {
	KC_STOP,
	KC_PLAY,
	KC_MSG,
	KC_WRONGC,
	KC_NEW,
	KC_PASS,
	KC_ERR,
	KC_LOAD,
	KC_SAVE,
	KC_HELP
} KeyboardCommand;

typedef struct {
	KeyboardCommand kc;
	int x;
	int y;
	char msg[256];
} KeyboardCommandStruct;

typedef enum {
	NC_OK,
	NC_ERR,
} NetCommand;

typedef struct {
	NetCommand nc;
	NetMessage msg;
	Game* g;
} NetCommandStruct;

// Sync. variables / data structures
pthread_t inputThread;
pthread_t netThread;
WINDOW* pisWin = NULL;
WINDOW* logWin = NULL;
WINDOW* inputWin = NULL;
pthread_mutex_t dataMutex;
pthread_cond_t dataCond;
Boolean condVariable;
Game* activeGame = NULL;

struct _syncData {
	Boolean fromNet;	// true - data from socket, false - from keyboard
	union {
		KeyboardCommandStruct keyb;
		NetCommandStruct net;
	} data;
} syncData;

/**
 * Ends curses library and terminates program
 *
 * @param retval Value to return
 */
static void exit_app(int retval)
{
	endwin();
	exit(retval);
}

/**
 * Refreshes all windows
 */
static void refreshAll()
{
  if (logWin == NULL || pisWin == NULL || inputWin == NULL) {
    return;
  }
	wrefresh(logWin);
	wrefresh(pisWin);
	wrefresh(inputWin);
}

/**
 * Prints log message into log window
 *
 * @param msg Message
 */
static void printlog(char* msg)
{
	waddstr(logWin, "! ");
	waddstr(logWin, msg);
	refreshAll();
}

/**
 * Prints message from opponent
 *
 * @param msg Message
 */
static void printmsgNet(char* msg)
{
	waddstr(logWin, ">> ");
	waddstr(logWin, msg);
	waddstr(logWin, "\n");
	refreshAll();
}

/**
 * Prints my message to opponent
 *
 * @param msg Message
 */
static void printmsgMe(char* msg)
{
	waddstr(logWin, "> ");
	waddstr(logWin, msg);
	waddstr(logWin, "\n");
	refreshAll();
}

/**
 * Prints help
 */
static void printHelp()
{
	printlog("msg : sends msg to the opponent\n");
	printlog("//a2 : plays at a 2\n");
	printlog("/n : starts new game with same parameters as previous\n");
	printlog("/l file : loads stored game\n");
	printlog("/s file : saves current game into file\n");
	printlog("/p : pass\n");
	printlog("/q : quit\n");
}

/**
 * Draws game into game window
 *
 * @param g Game to print
 */
static void printGame(Game* g)
{
	int i, j;

	wclear(pisWin);
	wprintw(pisWin, "O:%2d, X:%2d, *: %c, Y: %c\n", g->blueScore, g->redScore, (g->gs == GS_RED) ? 'x' : 'o', (g->whoAmI == GS_RED) ? 'x' : 'o');
	waddstr(pisWin, "   ");
	for (j = 0; j < g->cols; j++) {
		waddch(pisWin, 'a' + j);
		waddch(pisWin, ' ');
	}
	waddch(pisWin, '\n');
	for (j = 0; j < g->rows; j++) {
		wprintw(pisWin, "%2d ", j + 1);
		for (i = 0; i < g->cols; i++) {
			switch (g->fields[j][i]) {
				case FS_BLANK:
					waddstr(pisWin, "- ");
					break;
				case FS_RED:
					waddstr(pisWin, "x ");
					break;
				case FS_BLUE:
					waddstr(pisWin, "o ");
					break;
				case FS_REDX:
					waddstr(pisWin, "# ");
					break;
				case FS_BLUEX:
					waddstr(pisWin, "@ ");
					break;
			}
		}
		waddch(pisWin, '\n');
	}
	wrefresh(pisWin);
	wrefresh(inputWin);
}

/**
 * Processes command from keyboard
 *
 * @param s String from keyboard
 * @param x If s is a playat command, then this is filled with x coord.
 * @param y If s is a playat command, then this is filled with y coord.
 * @return Parsed command from keyboard
 */
static KeyboardCommand parseCommand(char* s, int* x, int* y)
{
	if (s[0] != '/') {
		return KC_MSG;
	}

	switch (s[1]) {
		case '/':
			if (strlen(s) < 4) {
				return KC_WRONGC;
			}
			*x = s[2] - 'a';
			*y = s[3] - '1';
			if (*y < 0 || *y > 99 || *x < 0 || *x > 99) {
				return KC_WRONGC;
			}

			if (s[4] >= '0' && s[4] <= '9') {
				*y = (*y + 1)*10 + s[4] - '0' - 1;
			}
			
			return KC_PLAY;
			break;
		case 'q':
			return KC_STOP;
			break;
		case 'n':
			return KC_NEW;
			break;
		case 'p':
			return KC_PASS;
			break;
		case 'h':
			return KC_HELP;
			break;
		case 'l':
			if (strlen(s) < 4) {
				return KC_WRONGC;
			}
			if (s[2] != ' ') {
				return KC_WRONGC;
			}
			return KC_LOAD;
		case 's':
			if (strlen(s) < 4) {
				return KC_WRONGC;
			}
			if (s[2] != ' ') {
				return KC_WRONGC;
			}
			return KC_SAVE;
		default:
			return KC_WRONGC;
	}
}

/**
 * Executes command from keyboard
 *
 * @param sockFd socket to opponent
 * @param g Game structure. May be freed and recreated
 * @param keyb Command from keyboard
 * @return TRUE - don't end program, FALSE - end program
 */
static Boolean executeCommandKeyb(int sockFd, Game** g, KeyboardCommandStruct* keyb)
{
	Game* newG;
	switch (keyb->kc) {
		case KC_NEW:
			newG = newGame((*g)->rows, (*g)->cols, (*g)->gs, (*g)->whoAmI, (*g)->pisLen);
			deleteGame(*g);
			*g = newG;
			if (!sendCommand(sockFd, NM_GAME)) {
				printlog("Error sending new game mesasge");
				return FALSE;
				break;
			}
			oppositeWhoAmI(*g);
			if (!storeGame(sockFd, *g)) {
				printlog("Error sending game");
				return FALSE;
			}
			oppositeWhoAmI(*g);
			printGame(*g);
			break;
		case KC_LOAD:
			newG = loadGameFromFile(keyb->msg);
			if (newG == NULL) {
				printlog("Error loading new game\n");
				break;
			} else {
				deleteGame(*g);
				*g = newG;
			}
			if (!sendCommand(sockFd, NM_GAME)) {
				printlog("Error sending new game mesasge\n");
				return FALSE;
				break;
			}
			oppositeWhoAmI(*g);
			if (!storeGame(sockFd, *g)) {
				printlog("Error sending game\n");
				return FALSE;
			}
			oppositeWhoAmI(*g);
			printGame(*g);
			break;
		case KC_SAVE:
			if (!saveGameToFile(keyb->msg, *g)) {
				printlog("Error saving game\n");
			}
			printlog("Game saved\n");
			break;
		case KC_STOP:
			if (!sendCommand(sockFd, NM_STOP)) {
				printlog("Error sending stop message\n");
			}
			return FALSE;
			break;
		case KC_MSG:
			if (!sendCommandMsg(sockFd, keyb->msg)) {
				printlog("Error sending message");
			}
			printmsgMe(keyb->msg);
			break;
		case KC_PLAY:
			if ((*g)->gs != (*g)->whoAmI) {
				printlog("It's not your turn\n");
				break;
			}
			if (!playAt(*g, (*g)->gs, keyb->x, keyb->y)) {
				printlog("Wrong move\n");
			} else {
				if (!sendCommandPlay(sockFd, keyb->x, keyb->y)) {
					printlog("Error sending message\n");
					return FALSE;
				} else {
					switchTurn(*g);
					printGame(*g);
				}
			}
			break;
		case KC_PASS:
			if ((*g)->gs != (*g)->whoAmI) {
				printlog("It's not your turn\n");
				break;
			}
			if (!sendCommand(sockFd, NM_PASS)) {
				perror("Error sending pass message");
				break;
			}
			switchTurn(*g);
			printGame(*g);
			break;
		case KC_HELP:
			printHelp();
			break;
		case KC_WRONGC:
			printlog("Wrong command\n");
			break;
		case KC_ERR:
			printlog("Error reading from keyboard\n");
			return FALSE;
			break;
	}
	return TRUE;
}

/**
 * Executes command from net
 *
 * @param sockFd socket to opponent
 * @param g Game structure. May be freed and recreated
 * @param keyb Command from keyboard
 * @return TRUE - don't end program, FALSE - end program
 */
static Boolean executeCommandNet(int sockFd, Game** g, NetCommandStruct* net)
{
	switch (net->msg.type) {
		case NM_PLAYAT:
			if (!playAt(*g, (*g)->gs, net->msg.x, net->msg.y)) {
				printlog("Error in synchronisation\n");
				return FALSE;
			} else {
				switchTurn(*g);
				printGame(*g);
			}
			break;
		case NM_MSG:
			net->msg.str[255] = '\0';
			printmsgNet(net->msg.str);
			refresh();
			break;
		case NM_STOP:
			return FALSE;
			break;
		case NM_GAME:
			deleteGame(*g);
			(*g) = net->g;
			printGame(*g);
			break;
		case NM_PASS:
			switchTurn(*g);
			printGame(*g);
			break;
		default:
			printlog("Unrecognised net message\n");
			return FALSE;
	}
	return TRUE;
}

/**
 * Function executed in a separated thread
 * waits for keyboard input and sends messages to the main thread
 *
 * @param isServV If this program acts as server
 * @return Slways NULL
 */
static void* inputFunc(void* isServerV)
{
	char buf[256];
	//Boolean isServer = *((Boolean*) isServerV);

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	while (TRUE) {
		if (wgetnstr(inputWin, buf, 255) == ERR) {
			if (errno == EINTR) {
				continue;
			}
			exit_app(2);
		}

		pthread_mutex_lock(&dataMutex);
		condVariable = TRUE;

		syncData.data.keyb.kc = parseCommand(buf, &syncData.data.keyb.x, &syncData.data.keyb.y);
		if (syncData.data.keyb.kc == KC_MSG) {
			strncpy(syncData.data.keyb.msg, buf, 255);
		} else if (syncData.data.keyb.kc == KC_LOAD || syncData.data.keyb.kc == KC_SAVE) {
			strncpy(syncData.data.keyb.msg, buf + 3, 252);
		}
		syncData.fromNet = FALSE;

		pthread_cond_signal(&dataCond);
		pthread_mutex_unlock(&dataMutex);
	}

	return NULL;
}

/**
 * Function executed in a separated thread
 * waits for commands from opponent from network and sends messages to the main thread
 *
 * @param sockFdV Pointer to socket on which to listen
 * @return Always NULL
 */
static void* netFunc(void* sockFdV)
{
	int sockFd = *((int*) sockFdV);
	struct pollfd inSet[1];
	int retVal;
	sigset_t signals;

 	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	memset(inSet, 0, sizeof(inSet));
	inSet[0].fd = sockFd;
	inSet[0].events = POLLIN;

	while (TRUE) {
		inSet[0].revents = 0;

		retVal = poll(inSet, 1, -1);

		if (retVal == -1) {
			if (errno == EINTR) {
				continue;
			}
			exit(2);
		}

		pthread_mutex_lock(&dataMutex);
		condVariable = TRUE;
		syncData.fromNet = TRUE;
		

		if (inSet[0].revents != POLLIN) {
/*			sigemptyset(&signals);
			sigpending(&signals);
			if (sigismember(&signals, SIGWINCH)) {
				printlog("wokno\n");
				break;
			}*/
			wprintw(logWin, "tohle: %d\n", inSet[0].revents);
			goto EOut;
			break;
		}

		// download the data

		if (!recvNetMessage(sockFd, &syncData.data.net.msg)) {
			goto EOut;
			break;
		}
		
		// game data follow
		if (syncData.data.net.msg.type == NM_GAME) {
			if ((syncData.data.net.g = readGame(sockFd, TRUE)) == NULL) {
				goto EOut;
			}

		}
		
		syncData.data.net.nc = NC_OK;
		pthread_cond_signal(&dataCond);
		pthread_mutex_unlock(&dataMutex);

		if (syncData.data.net.msg.type == NM_STOP) {
			break;
		}
		
	}
	return NULL;

EOut:
	syncData.data.net.nc = NC_ERR;
	pthread_cond_signal(&dataCond);
	pthread_mutex_unlock(&dataMutex);
	return NULL;
}

/**
 * Main game loop. waits for messages from inputThread and netThread
 *
 * @param g First game being played
 * @param sockFd Socket to opponent
 * @param isServer If this program acts as server
 * @return TRUE - successfull, FALSE - otherwise
 */
static Boolean gameLoop(Game* g, int sockFd, Boolean isServer)
{
	Boolean play = TRUE;
	Boolean retVal = TRUE;

	condVariable = FALSE;
	pthread_mutex_init(&dataMutex, NULL);
	pthread_mutex_lock(&dataMutex);

	if (pthread_cond_init(&dataCond, NULL) != 0) {
		return FALSE;
	}
	if (pthread_create(&inputThread, NULL, inputFunc, &isServer) != 0) {
		return FALSE;
	}
	if (pthread_create(&netThread, NULL, netFunc, &sockFd) != 0) {
		pthread_cancel(inputThread);
		pthread_detach(inputThread);
		return FALSE;
	}

	while (play) {

		while (!condVariable) {
			pthread_cond_wait(&dataCond, &dataMutex);
		}
		condVariable = FALSE;
		// data from net
		if (syncData.fromNet) {
			if (syncData.data.net.nc == NC_ERR) {
				retVal = FALSE;
				break;
			}

			if (!executeCommandNet(sockFd, &g, &(syncData.data.net))) {
				play = FALSE;
			}
			
		// data from keyboard
		} else {
			if (syncData.data.keyb.kc == KC_ERR) {
				retVal = FALSE;
				break;
			}
 			if (!executeCommandKeyb(sockFd, &g, &(syncData.data.keyb))) {
 				play = FALSE;
 			}
		}
	}

	pthread_cancel(inputThread);
	pthread_detach(inputThread);
	pthread_cancel(netThread);
	pthread_detach(netThread);

	return retVal;
}

/**
 * Prints help about using of this program
 */
static void showHelp()
{
	printf("netpis\n");
	printf("parameters:\n");
	printf("-s : server\n");
	printf("-c ip/hostname : client\n");
	printf("-p port : port where to listen, or to which to connect\n");
	printf("-x size : initial width of game field\n");
	printf("-y size : initial height of game field\n");
	printf("-l size : initial length of piskvorek\n");
	printf("-f file : start with stored in file\n");
	printf("-b : server has x\n");
	printf("-i : client starts\n");
	printf("-h : show this help\n");
	printf("\nexample:\n");
	printf("server: netpis -s -p 10000 -b -l 3\n");
	printf("client: netpis -c 192.168.0.1 -p 10000\n");
	printf("\nFor help during the game, type /h\n");
	exit(0);
}


/**
 * Helper routine for main()
 */
static void copyArg(char** where, char* what)
{
	if (!(*where = (char*)malloc(strlen(what) + 1))) {
		perror(NULL);
		exit(1);
	}
	strcpy(*where, what);
}

static void exit_signal(int signum)
{
//	exit_app(1);
}

/**
 * Creates curses windows
 *
 * @return TRUE - ok, FALSE - ko
 */
Boolean createWindows()
{
	pisWin = newwin(LINES / 2, COLS / 2, 0, 0);
	logWin = newwin(LINES, COLS / 2, 0, COLS / 2);
	inputWin = newwin(LINES / 2, COLS / 2, LINES / 2, 0);

	if (pisWin == NULL || logWin == NULL || inputWin == NULL) {
		return FALSE;
	}

	scrollok(logWin, TRUE);
	idlok(logWin, TRUE);
	scrollok(inputWin, TRUE);
	idlok(inputWin, TRUE);

	return TRUE;
}

/**
 * Initializes curses library
 *
 * @return TRUE - ok, FALSE - unsuccessful
 */
Boolean initScreen()
{
	initscr();
	cbreak();
//	raw();
	echo();
	nl();
	clear();
	if (!createWindows()) {
		return FALSE;
	}
	refresh();
	return TRUE;
}


void resize(int signum)
{
  mvwin(pisWin, 0, 0);
  mvwin(logWin, 0, COLS / 2);
  mvwin(inputWin, LINES / 2, 0);
  wresize(pisWin, LINES / 2, COLS / 2);
  wresize(logWin, LINES, COLS / 2);
  wresize(inputWin, LINES / 2, COLS / 2);

  if (activeGame != NULL) {
    printGame(activeGame);
    refreshAll();
  }
}


/**
 * Main()
 */
int main(int argc, char* argv[])
{
	int opt;
	int servSock, clientSock;
	Game* newG;
	struct sigaction act;
	struct fArgs {
		char* sClient;
		char* sFile;
		int port;
		int xsize;
		int ysize;
		int len;
		Boolean fServerCross;
		Boolean fClientStarts;
		Boolean	fServer;
		Boolean fClient;
		Boolean fFile;
		Boolean fIp;
		Boolean fX;
		Boolean fY;
		Boolean fLen;
		Boolean fHelp;
	} args;

	// Signal handling
	act.sa_handler = resize;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	sigaction(SIGWINCH, &act, NULL);

	act.sa_handler = exit_signal;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	sigaction(SIGINT, &act, NULL);

	// Parameters
	memset(&args, 0, sizeof(args));
	args.port = DEF_PORT;
	args.xsize = DEF_XSIZE;
	args.ysize = DEF_YSIZE;
	while ( (opt = getopt(argc, argv ,":sc:p:x:y:l:f:hbi") ) !=-1 ) {
		switch (opt) {
			case 's':
				args.fServer = TRUE;
				break;
			case 'c':
				args.fClient = TRUE;
				copyArg(&args.sClient, optarg);
				break;
			case 'f':
				args.fFile = TRUE;
				copyArg(&args.sFile, optarg);
				break;
			case 'x':
				args.fX = TRUE;
				args.xsize = atoi(optarg);
				break;
			case 'y':
				args.fY = TRUE;
				args.ysize = atoi(optarg);
				break;
			case 'l':
				args.fLen = TRUE;
				args.len = atoi(optarg);
				break;
			case 'h':
				args.fHelp = TRUE;
				break;
			case 'b':
				args.fServerCross = TRUE;
				break;
			case 'i':
				args.fClientStarts = TRUE;
				break;
			case 'p':
				args.port = atoi(optarg);
				break;
			case '?':
			case ':':
				args.fHelp = TRUE;
				break;
			default:	// unreachable
				args.fHelp = TRUE;
				break;
		}
	}

	if (args.fHelp) {
		showHelp();
	}

	if ((args.fServer && args.fClient) || !(args.fServer || args.fClient)) {
		showHelp();
	}
	if (args.fFile && (args.fServerCross || args.fClientStarts)) {
		showHelp();
	}

	if ((args.fX && (args.xsize <= 0 || args.xsize > MAX_XSIZE)) || (args.fY && (args.ysize <= 0 || args.ysize > MAX_YSIZE)) ||
		(args.fLen && (args.len >= args.xsize || args.len >= args.ysize || args.len <= MIN_PISLEN))) {

		showHelp();
	}

	// Initialize curses library
	if (!initScreen()) {
		printf("unable to init curses\n");
		exit_app(1);
	}


	// RED = cross, BLUE = circle
	// SERVER
	if (args.fServer) {
		if (!args.fFile) {
			newG = newGame(
				(args.fY) ? args.ysize : DEF_YSIZE,
				(args.fX) ? args.xsize : DEF_XSIZE,
				((args.fServerCross && !args.fClientStarts) || (!args.fServerCross && args.fClientStarts)) ? GS_RED : GS_BLUE,
				(args.fServerCross) ? GS_RED : GS_BLUE,
				(args.fLen) ? args.len : DEF_LEN
				);
		} else {
			newG = loadGameFromFile(args.sFile);
		}

		if (!newG) {
			printlog("Error loading or creating new game\n");
			exit_app(1);
		}

		if ((servSock = serverAccept(args.port)) == -1) {
			printlog("Error in waiting for client\n");
			exit_app(1);
		}

		oppositeWhoAmI(newG);
		if (!storeGame(servSock, newG)) {
			close(servSock);
			printlog("Error sending file to client\n");
			exit_app(1);
		}
		oppositeWhoAmI(newG);

		activeGame = newG;
    printGame(newG);
		gameLoop(newG, servSock, TRUE);

	// CLIENT
	} else {
		if ((clientSock = clientConnect(args.sClient, args.port)) == -1) {
			printlog("Error connecting to server\n");
			exit_app(1);
		}
		if ((newG = readGame(clientSock, TRUE)) == NULL) {
			close(clientSock);
			printlog("Error receiving game from server\n");
			exit_app(1);
		}

		if (args.fFile) {
			deleteGame(newG);
			newG = loadGameFromFile(args.sFile);
			if (newG == NULL) {
				printlog("Error loading new game\n");
				exit_app(1);
			}
			if (!sendCommand(clientSock, NM_GAME)) {
				printlog("Error sending new game mesasge\n");
				exit_app(1);
			}
			oppositeWhoAmI(newG);
			if (!storeGame(clientSock, newG)) {
				printlog("Error sending game\n");
				exit_app(1);
			}
			oppositeWhoAmI(newG);
		}

    activeGame = newG;
		printGame(newG);
		gameLoop(newG, clientSock, FALSE);
		close(clientSock);
	}

	exit_app(0);

	return 0;
}

